from typing import TYPE_CHECKING, Any, Dict, List, Union
from typing_extensions import Literal
from pydavinci.utils import is_resolve_obj
from pydavinci.wrappers.marker import MarkerCollection
from pydavinci.wrappers.mediapoolitem import MediaPoolItem

if TYPE_CHECKING:
    from pydavinci.wrappers._resolve_stubs import PyRemoteTimelineItem  # type: ignore


class TimelineItem(object):
    def __init__(self, obj: "PyRemoteTimelineItem") -> None:

        if is_resolve_obj(obj):
            self._obj: "PyRemoteTimelineItem" = obj
        else:
            raise TypeError(f"{type(obj)} is not a valid {self.__class__.__name__} type")

        self.markers = MarkerCollection(self)

    @property
    def name(self) -> str:
        """
        Gets name of ``TimelineItem``

        Returns:
            str: name
        """
        return self._obj.GetName()

    @property
    def duration(self) -> int:
        """
        Get duration in frames

        Returns:
            int: duration in frames
        """
        return self._obj.GetDuration()

    @property
    def start(self) -> int:
        """
        Returns the start frame position on the timeline.

        Returns:
            int: start frame position
        """
        return self._obj.GetStart()

    @property
    def end(self) -> int:
        """
        Returns the start frame position on the timeline.

        Returns:
            int: end frame position
        """
        return self._obj.GetEnd()

    @property
    def left_offset(self) -> int:
        """
        Returns the maximum extension by frame for clip from left side.

        Returns:
            int: left offset frame count
        """
        return self._obj.GetLeftOffset()

    @property
    def right_offset(self) -> int:
        """
        Returns the maximum extension by frame for clip from right side.

        Returns:
            int: right offset frame count
        """
        return self._obj.GetRightOffset()

    @property
    def properties(self) -> Dict[Any, Any]:
        """
        Gets all clip properties

        Returns:
            dict: dict with clip properties
        """

        return self._obj.GetProperty()

    def set_property(self, key: str, value: Union[str, int, float]) -> bool:
        """
        Sets property

        Args:
            name (str): property name
            value (Union[str, int, float]): property value

        Returns:
            bool: ``True`` if successful, ``False`` otherwise
        """

        return self._obj.SetProperty(key, value)

    def add_flag(self, color: str) -> bool:
        """
        Adds flag

        Args:
            color (str): valid flag

        Returns:
            bool: ``True`` if successful, ``False`` otherwise
        """
        return self._obj.AddFlag(color)

    @property
    def flags(self) -> List[str]:
        """
        Gets flag list

        Returns:
            List[str]: list of flags
        """
        return self._obj.GetFlagList()

    def clear_flags(self, color: str = "All") -> bool:
        """
        Clears all flags

        Args:
            color (str, optional): clears flags by ``color``. If `All` provided, clear all flags. Defaults to "All".

        Returns:
            bool: _description_
        """
        return self._obj.ClearFlags(color)

    @property
    def color(self) -> str:
        """
        Gets or sets clip color

        Args:
            color (str): color to be applied

        Returns:
            bool: ``True`` if successful, ``False`` otherwise
        """
        return self._obj.GetClipColor()

    @color.setter
    def color(self, color: str) -> bool:
        """
        Sets clip color

        Args:
            color (str): color to be applied

        Returns:
            bool: ``True`` if successful, ``False`` otherwise
        """
        return self._obj.SetClipColor(color)

    def clear_color(self) -> bool:
        """
        Clears clip color

        Returns:
            bool: ``True`` if successful, ``False`` otherwise
        """
        return self._obj.ClearClipColor()

        # TODO: Document these

    def add_color_version(self, name: str, type: Literal["local", "remote"]) -> bool:
        """Adds color version to this `TimelineItem`

        Args:
            name (str): version name
            type (Literal['local', 'remote']): whether to add a local or remote color version

        Returns:
            bool: ``True`` if successful, ``False`` otherwise
        """
        if type == "local":
            return self._obj.AddVersion(name, 0)
        elif type == "remote":
            return self._obj.AddVersion(name, 1)
        else:
            return False

    def delete_color_version(self, name: str, type: Literal["local", "remote"]) -> bool:
        """Adds color version from this `TimelineItem`

        Args:
            name (str): version name
            type (Literal['local', 'remote']): whether to delete a local or remote color version

        Returns:
            bool: ``True`` if successful, ``False`` otherwise
        """
        if type == "local":
            return self._obj.DeleteVersionByName(name, 0)
        elif type == "remote":
            return self._obj.DeleteVersionByName(name, 1)
        else:
            return False

    def load_color_version(self, name: str, type: Literal["local", "remote"]) -> bool:
        """Loads color version from `TimelineItem` named `name`

        Args:
            name (str): version name
            type (Literal['local', 'remote']): whether to load a local or remote color version

        Returns:
            bool: ``True`` if successful, ``False`` otherwise
        """
        if type == "local":
            return self._obj.LoadVersionByName(name, 0)
        elif type == "remote":
            return self._obj.LoadVersionByName(name, 1)
        else:
            return False

    def rename_version(self, oldname: str, newname: str, type: Literal["local", "remote"]) -> bool:
        """Renames a color version named `oldname` to `newname` on this `TimelineItem`

        Args:
            oldname (str): current version name
            newname (str): new version name
            type (Literal['local', 'remote']): whether to rename a local or remote color version

        Returns:
            bool: ``True`` if successful, ``False`` otherwise
        """
        if type == "local":
            return self._obj.RenameVersionByName(oldname, newname, 0)
        elif type == "remote":
            return self._obj.RenameVersionByName(oldname, newname, 1)
        else:
            return False

    @property
    def mediapoolitem(self) -> "MediaPoolItem":
        """
        Returns the corresponding ``MediaPoolItem`` for this ``TimelineItem``

        Returns:
            MediaPoolItem: ``MediaPoolItem``
        """
        return MediaPoolItem(self._obj.GetMediaPoolItem())

    # /TODO: Add Stero and Fusion wrappers

    def set_lut(self, node_index: int, lut_path: str) -> bool:
        """
        Sets lut located on ``lut_path`` at ``node_index``

        Args:
            node_index (int): node index
            lut_path (str): lut path

        Returns:
            bool: ``True`` if successful, ``False`` otherwise
        """
        return self._obj.SetLUT(node_index, lut_path)

    def set_cdl(self, cdl: Dict[Any, Any]) -> bool:
        """
        Sets CDL

        Args:
            cdl (dict): valid CDL dict

        Example:
            ```python
            myclip.set_cdl({"NodeIndex" : "1",
            "Slope" : "0.5 0.4 0.2",
            "Offset" : "0.4 0.3 0.2",
            "Power" : "0.6 0.7 0.8",
            "Saturation" : "0.65"})
            ```

        Returns:
            bool: ``True`` if successful, ``False`` otherwise
        """
        return self._obj.SetCDL(cdl)

    def add_take(
        self, mediapool_item: "MediaPoolItem", startframe: int = 0, endframe: int = 0
    ) -> bool:
        """
        Adds ``mediapool_item`` as a new take. Initializes a take selector for the timeline item if needed.
        By default, the full clip extents is added. ``startframe`` and ``endFrame`` are
        optional arguments used to specify the extents.

        Args:
            mediapool_item (MediaPoolItem): media pool item to add as take
            startframe (int, optional): start frame for new take. Defaults to 0.
            endframe (int, optional): end frame for new take. Defaults to 0.

        Returns:
            bool: _description_
        """
        return self._obj.AddTake(mediapool_item._obj, startframe, endframe)

    @property
    def take(self) -> int:
        """
        Gets or sets current take

        Args:
            takeindex (int): take index for selection
        """
        return self._obj.GetSelectedTakeIndex()

    @take.setter
    def take(self, takeindex: int) -> None:
        """
        Sets selected take

        Args:
            takeindex (int): take index for selection

        Returns:
            bool: ``True`` if successful, ``False`` otherwise
        """
        self._obj.SelectTakeByIndex(takeindex)

    @property
    def takes(self) -> int:
        """
        Get total number of takes

        Returns:
            int: total number of takes
        """
        return self._obj.GetTakesCount()

    def take_info(self, takeindex: int = 0) -> Dict[Any, Any]:
        """
        Gets take info. If no ``takeindex`` provided, uses current selected take.

        Args:
            takeindex (int, optional): take index to get info from. Defaults to current take.

        Returns:
            dict: _description_
        """
        if takeindex:
            return self._obj.GetTakeByIndex(takeindex)
        else:
            return self._obj.GetTakeByIndex(self._obj.GetSelectedTakeIndex())

    def delete_take(self, takeindex: int) -> bool:
        """
        Deletes take index with index ``takeindex``

        Args:
            takeindex (int): take index to be deleted

        Returns:
            bool: ``True`` if successful, ``False`` otherwise
        """
        return self._obj.DeleteTakeByIndex(takeindex)

    def finalize_take(self, takeindex: int = 0) -> bool:
        """
        Finalizes take at ``takeindex``. If no ``takeindex`` provided, finalizes current take.

        Args:
            takeindex (int, optional): take index to be finalized. Defaults to current take.

        Returns:
            bool: _description_
        """
        if takeindex:
            self.take = takeindex
            return self._obj.FinalizeTake()
        else:
            return self._obj.FinalizeTake()

    def copy_grade_to(self, timeline_items: List["TimelineItem"]) -> bool:
        """
        Copies the current grade to all the items in ``timneline_items`` list

        Args:
            timeline_items (list): list with clips for the grade to be applied

        Returns:
            bool: ``True`` if successful, ``False`` otherwise
        """
        return self._obj.CopyGrades([x._obj for x in timeline_items])

    def __repr__(self) -> str:
        clip_repr = str(self.start) + "{| " + str(self.duration) + " |}" + str(self.end)
        return f'TimelineItem(Name:"{self.name}", In/Duration/Out: {clip_repr})'



from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union

from pydavinci.exceptions import ObjectNotFound
from pydavinci.main import resolve_obj
from pydavinci.utils import is_resolve_obj
from pydavinci.wrappers.settings.constructor import get_prj_settings

if TYPE_CHECKING:
    from pydavinci.wrappers._resolve_stubs import PyRemoteProject
    from pydavinci.wrappers.mediapool import MediaPool
    from pydavinci.wrappers.settings.constructor import ProjectSettings
    from pydavinci.wrappers.timeline import Timeline


class Project(object):
    def __init__(self, *args: Any) -> None:
        if args:
            if is_resolve_obj(args[0]):
                self._obj: "PyRemoteProject" = args[0]
            else:
                raise TypeError(f"{type(args[0])} is not a valid {self.__class__.__name__} type")
        else:
            self._obj = resolve_obj.GetProjectManager().GetCurrentProject()

        self._settings: Optional["ProjectSettings"] = None

    @property
    def settings(self) -> "ProjectSettings":
        """Returns the [`ProjectSettings`](../settings/project) interface."""

        if self._settings is None:
            self._settings = get_prj_settings(self)
            return self._settings

        else:
            return self._settings

    @property
    def mediapool(self) -> "MediaPool":
        """Returns the
        [``MediaPool``][pydavinci.wrappers.mediapool.MediaPool-attributes] object.
        Returns:
            (MediaPool): Media Storage object
        """
        from pydavinci.wrappers.mediapool import MediaPool

        return MediaPool()

    @property
    def timeline_count(self) -> int:
        """
        Get total timeline count on current project

        Returns:
            timeline count
        """
        return self._obj.GetTimelineCount()

    @property
    def name(self) -> str:
        """
        Gets or sets current project name

        Returns:
            project name
        """
        return self._obj.GetName()

    @name.setter
    def name(self, name: str) -> bool:
        """
        Project name. Use to rename project

        ``Project.name = 'New Project Name'``

        Returns:
            bool: ``True`` if successful, ``False`` otherwise
        """
        return self._obj.SetName(name)

    @property
    def presets(self) -> List[str]:
        """Gets a list of available project presets

        Returns:
            project presets
        """
        return self._obj.GetPresetList()

    def set_preset(self, preset_name: str) -> bool:
        """
        Activates ``preset_name``

        Args:
            preset_name (str): preset name

        Returns:
            bool: ``True`` if successful, ``False`` otherwise
        """
        return self._obj.SetPreset(preset_name)

    def add_renderjob(self, block: bool = True) -> str:
        """
        Adds current render settings to a render job.

        If there are already rendered jobs in the render queue and you're executing a lot of commands, there's a bug on
        the Davinci API that there's a chance it will return an empty string instead of the job ID.

        ``block`` blocks the program until we get a job id back from Davinci Resolve. It's ``True`` by default.

        Returns:
            str: render job id
        """
        job_id = self._obj.AddRenderJob()

        if block and job_id == "":
            while job_id == "":

                job_id = self._obj.AddRenderJob()

            return job_id

        return job_id

    def delete_renderjob(self, job_id: str) -> bool:
        """
        Deletes render job ``job_id``

        Args:
            job_id (str): render job id

        Returns:
            bool: ``True`` if successful, ``False`` otherwise
        """
        return self._obj.DeleteRenderJob(job_id)

    def delete_all_renderjobs(self) -> bool:
        """
        Deletes all renderjobs

        Returns:
            bool: ``True`` if successful, ``False`` otherwise
        """
        return self._obj.DeleteAllRenderJobs()

    @property
    def render_jobs(self) -> List[str]:
        """
        Gets current list of render jobs

        Returns:
            list: render job list
        """
        return self._obj.GetRenderJobList()

    @property
    def render_presets(self) -> List[str]:
        """
        Gets available __render__ presets list

        Returns:
            list: available render presets list
        """
        return self._obj.GetRenderPresetList()

    def render(self, job_ids: Optional[List[str]] = None, interactive: bool = True) -> bool:
        """
        Render jobs

        Args:
            job_ids (Union[None, List[str]], optional): Renders provided list of ``job ids``. If ``None`` provided, render all jobs in queue. Defaults to ``None``.
            interactive (bool, optional): Whether to use interactive mode. When set to ``True``, enables error feedback in the UI during rendering.

        Returns:
            bool: ``True`` if successful, ``False`` otherwise
        """
        # TODO: the API supports passing only jobs ids: str as arguments. Implement here
        # using *args and update stub.
        if not job_ids:
            return self._obj.StartRendering(isInteractiveMode=interactive)
        else:
            return self._obj.StartRendering(job_ids, isInteractiveMode=interactive)

    def stop_render(self) -> None:
        """
        Stops all rendering.

        Returns:
            None: None
        """
        return self._obj.StopRendering()

    def render_status(self, job_id: str) -> Dict[Any, Union[str, int]]:
        """
        Gets render status on ``job_id``

        Args:
            job_id (str): job id

        Returns:
            dict: dictionary with render status

        Render Status:
            The dictionary returned looks like this when rendering:
            ```python
            {'JobStatus': 'Rendering',
            'CompletionPercentage': 92,
            'EstimatedTimeRemainingInMs': 1000}
            ```
            And like this when the render on the provided job id is complete:
            ```python
            {'JobStatus': 'Complete',
            'CompletionPercentage': 100,
            'TimeTakenToRenderInMs': 25991}
            ```
        """
        return self._obj.GetRenderJobStatus(job_id)

    @property
    def render_formats(self) -> Dict[Any, Any]:
        """
        Gets all possible render formats.

        Returns:
            dict: dict with render formats
        """
        return self._obj.GetRenderFormats()

    def get_render_codecs(self, render_format: str) -> Dict[Any, Any]:
        """Returns all possible render codecs.

        Args:
            render_format (str): render format

        Returns:
            Dict[Any, Any]: render codecs
        """
        return self._obj.GetRenderCodecs(render_format)

    @property
    def current_render_format_and_codec(self) -> Dict[Any, Any]:
        """
        Gets current render format and codec.

        Returns:
            dict: dict with current render format and codec
        """
        return self._obj.GetCurrentRenderFormatAndCodec()

    def set_render_format_and_codec(self, format: str, codec: str) -> bool:
        """
        Sets current ``format`` and ``codec``


        Args:
            format (str): render format
            codec (str): render codec

        Returns:
            bool: ``True`` if successful, ``False`` otherwise
        """
        return self._obj.SetCurrentRenderFormatAndCodec(format, codec)

    @property
    def render_mode(self) -> str:
        """Gets or sets current render mode. ``single`` for single clip and ``individual`` for individual clips.

        Returns:
            render mode ``single`` or ``individual``
        """
        if self._obj.GetCurrentRenderMode() == 1:
            return "single"
        else:
            return "individual"

    @render_mode.setter
    def render_mode(self, mode: str) -> bool:
        if "individual" in mode:
            return self._obj.SetCurrentRenderMode(0)
        elif "single" in mode:
            return self._obj.SetCurrentRenderMode(1)
        else:
            raise ValueError(
                'Render mode must be "single" or "individual", \
                for single clip and individual clips, respectively.'
            )

    def available_resolutions(
        self, format: Optional[str] = None, codec: Optional[str] = None
    ) -> List[Dict[Any, Any]]:
        """Returns list of resolutions applicable for the given render ``format`` and render ``codec``.

        Returns full list of resolutions if no argument is provided.
        Each element in the list is a dictionary with 2 keys "Width" and "Height".

        Args:
            format (str): render format
            codec (str): render codec

        Returns:
            List of available resolutions
        """
        if format and codec is None:
            return self._obj.GetRenderResolutions()
        else:
            return self._obj.GetRenderResolutions(format, codec)

    def is_rendering(self) -> bool:
        """
        Checks if DaVinci Resolve is rendering.

        Returns:
            bool: `True` if rendering, `False` otherwise
        """
        return self._obj.IsRenderingInProgress()

    def load_render_preset(self, preset_name: str) -> bool:
        """
        Loads render preset ``preset_name``.

        Args:
            preset_name (str): preset name

        Returns:
            bool: ``True`` if successful, ``False`` otherwise
        """
        return self._obj.LoadRenderPreset(preset_name)

    def save_render_preset_as(self, preset_name: str) -> bool:
        """
        Save current preset as ``preset_name``.

        Args:
            preset_name (str): preset name

        Returns:
            bool: ``True`` if successful, ``False`` otherwise
        """
        return self._obj.SaveAsNewRenderPreset(preset_name)

    def set_render_settings(self, render_settings: Dict[Any, Any]) -> bool:
        """
        Set render settings.

        Render Settings:
        ```python
            render_settings = {
            "SelectAllFrames": bool, # (when set True, the settings MarkIn and MarkOut are ignored)
            "MarkIn": int,
            "MarkOut": int,
            "TargetDir": str,
            "CustomName": str,
            "UniqueFilenameStyle": int, # 0 - Prefix, 1 - Suffix.
            "ExportVideo": bool,
            "ExportAudio": bool,
            "FormatWidth": int,
            "FormatHeight": int,
            "FrameRate": float, # (examples: 23.976, 24)
            "PixelAspectRatio": str, # (for SD resolution: "16_9" or "4_3") (other resolutions: "square" or "cinemascope")
            "VideoQuality": Union[int, str],
               # possible values for current codec (if applicable):
               # 0 (int) - will set quality to automatic
               # [1 -> MAX] (int) - will set input bit rate
               # ["Least", "Low", "Medium", "High", "Best"] (String) - will set input quality level
            "AudioCodec": str, # (example: "aac")
            "AudioBitDepth": int,
            "AudioSampleRate": int,
            "ColorSpaceTag" : str, # (example: "Same as Project", "AstroDesign")
            "GammaTag" : str, # (example: "Same as Project", "ACEScct")
            "ExportAlpha": bool,
            "EncodingProfile": str, # (example: "Main10"). Can only be set for H.264 and H.265.
            "MultiPassEncode": bool, # Can only be set for H.264.
            "AlphaMode": int, # 0 - Premultiplied, 1 - Straight. Can only be set if "ExportAlpha" is true.
            "NetworkOptimization": bool, # Only supported by QuickTime and MP4 formats.
        ```
        Args:
            render_settings (dict): dictionary with render settings

        Returns:
            bool: ``True`` if successful, ``False`` otherwise
        """
        return self._obj.SetRenderSettings(render_settings)

    def get_setting(self, settingname: Optional[str] = None) -> Any:
        """
        Get project setting. If no setting provided, returns a dict with all settings.

        Args:
            setting (str): setting name

        Returns:
            dict: dict with setting name and value
        """
        if settingname:
            return self._obj.GetSetting(settingname)
        return self._obj.GetSetting()

    def set_setting(self, setting: str, value: Any) -> bool:
        """
        Set project setting.

        Args:
            setting (str): setting name
            value (Any): setting value

        Returns:
            bool: ``True`` if successful, ``False`` otherwise
        """
        return self._obj.SetSetting(setting, value)

    def save(self) -> bool:
        """
        Saves project.

        Returns:
            bool: ``True`` if successful, ``False`` otherwise
        """
        return resolve_obj.GetProjectManager().SaveProject()

    ## note: after loading the first project, it's impossible
    ## to close a project and go back to the window with only
    ## the project manager showing.
    ## It defaults to a 'Untitled Project', and you can't close it
    def close(self) -> bool:
        """
        Closes current project.

        Returns:
            bool: ``True`` if successful, ``False`` otherwise
        """
        return resolve_obj.GetProjectManager().CloseProject(self._obj)

    def open_timeline(self, name: str) -> bool:
        """
        Opens timeline named ``name``.

        Args:
            name (str): timeline name

        Returns:
            bool: ``True`` if successful, ``False`` otherwise
        """
        from pydavinci.wrappers.timeline import Timeline

        count = self._obj.GetTimelineCount()
        for i in range(count):
            tl = Timeline(self._obj.GetTimelineByIndex(i + 1))
            if tl.name == name:
                return self._obj.SetCurrentTimeline(tl._obj)

        raise ObjectNotFound("Couldn't find timeline by name.")

    @property
    def timeline(self) -> "Timeline":
        """
        Returns current [``Timeline``][pydavinci.wrappers.timeline.Timeline-attributes] object

        Returns:
            Timeline(Timeline): timeline object
        """
        from pydavinci.wrappers.timeline import Timeline

        return Timeline(self._obj.GetCurrentTimeline())

    def refresh_luts(self) -> bool:
        """
        Refresh luts.

        Returns:
            None: None
        """
        return self._obj.RefreshLUTList()

    def __repr__(self) -> str:
        return f'Project(Name: "{self.name})"'



from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union

from pydavinci.utils import is_resolve_obj
from pydavinci.wrappers.marker import MarkerCollection

if TYPE_CHECKING:
    from pydavinci.wrappers._resolve_stubs import PyRemoteMediaPoolItem


class MediaPoolItem(object):
    # TODO:
    # Implement a way to acess metadata such as mediapoolitem.metadata['Good Take'] = True
    # Meed to mess around with a private dict that uses
    # the SetMetadata() when internal dict updates

    def __init__(self, obj: "PyRemoteMediaPoolItem") -> None:
        if is_resolve_obj(obj):
            self._obj: "PyRemoteMediaPoolItem" = obj
        else:
            raise TypeError(f"{type(obj)} is not a valid {self.__class__.__name__} type")

        self.markers = MarkerCollection(self)

    @property
    def name(self) -> str:
        """
        Returns:
            (str): ``MediaPoolItem`` name
        """
        return self._obj.GetName()

    def get_metadata(self, metadata_key: Optional[str] = None) -> Union[str, Dict[Any, Any]]:
        """Gets metadata ``metadata_key`` for ``MediaPoolItem``. If no ``metadata_key`` is provided,
           returns a ``Dict`` with all available metadata. Can return an empty ``dict`` if there's no metadata.

        Args:
            metadata_key (Optional[Any]): metadata key

        Returns:
            (Union[str, Dict[Any, Any]]): ``Dict`` or ``str`` corresponding to ``metadata_key``
        """
        if metadata_key:
            return self._obj.GetMetadata(metadata_key)
        return self._obj.GetMetadata()

    def set_metadata(self, meta_dict: Any) -> bool:
        """
        Sets metadata with ``meta_dict``:
        ```python
            meta_dict = {
                metadata_key: metadata value
                }
        ```

        It's recommended you validate which metadata you wan't to change first
        by using [MediaPoolItem.get_metadata()][pydavinci.wrappers.mediapoolitem.MediaPoolItem.get_metadata]
        and getting a dict with all the metadata to see which one you want to alter.

        This will probably change for the better for version 1.0

        Args:
            meta_dict (dict): dict with metadata to be set

        Returns:
            bool: ``True`` if successful, ``False`` otherwise
        """
        return self._obj.SetMetadata(meta_dict)

    @property
    def media_id(self) -> str:
        """
        Returns:
            (str): ``MediaPoolItem`` UUID
        """
        return self._obj.GetMediaId()

    def add_flag(self, color: str) -> bool:
        """
        Adds a flag

        Args:
            color (str): flag color

        Returns:
            bool: ``True`` if successful, ``False`` otherwise
        """
        return self._obj.AddFlag(color)

    @property
    def flags(self) -> List[str]:
        """
        Gets flag list

        Returns:
            list of valid flag colors
        """
        return self._obj.GetFlagList()

    def clear_flags(self, color: str = "All") -> bool:
        """
        Clears flags

        Args:
            color (str, optional): Clears flag by ``color``. If none provided, defaults to "All" which clears all flags.

        Returns:
            ``True`` if successful, ``False`` otherwise
        """
        return self._obj.ClearFlags(color)

    @property
    def color(self) -> str:
        """
        Gets or sets clip color

        Args:
            color (str): new clip color

        Returns:
            clip color
        """
        return self._obj.GetClipColor()

    @color.setter
    def color(self, color: str) -> bool:
        """
        Sets clip color

        Args:
            color (str): new clip color

        Returns:
            bool: _description_
        """
        return self._obj.SetClipColor(color)

    def clear_color(self) -> bool:
        """
        Clears clip color

        Returns:
            bool: ``True`` if successful, ``False`` otherwise
        """
        return self._obj.ClearClipColor()

    @property
    def properties(self) -> Union[str, Dict[Any, Any]]:
        """
        Gets all clip properties

        Returns:
            dict: dict with clip properties
        """
        return self._obj.GetClipProperty()

    def set_property(self, name: str, value: Union[str, int, float]) -> bool:
        """
        Sets property

        Args:
            name (str): property name
            value (Union[str, int, float]): property value

        Returns:
            bool: ``True`` if successful, ``False`` otherwise
        """
        return self._obj.SetClipProperty(name, value)

    def link_proxy(self, path: str) -> bool:
        """
        Links media located at ``path`` to this ``MediaPoolItem``

        Args:
            path (str): _absolute_ path to proxy media

        Returns:
            bool: ``True`` if successful, ``False`` otherwise
        """
        return self._obj.LinkProxyMedia(path)

    def unlink_proxy(self) -> bool:
        """
        Unlinks proxy media of this ``MediaPoolItem``

        Returns:
            bool: ``True`` if successful, ``False`` otherwise
        """
        return self._obj.UnlinkProxyMedia()

    def replace_clip(self, path: str) -> bool:
        """
        Replaces the underlying asset and metadata of ``MediaPoolItem`` with the specified absolute clip path.

        Args:
            path (str): path to clip

        Returns:
            bool: ``True`` if successful, ``False`` otherwise
        """
        return self._obj.ReplaceClip(path)

    def __repr__(self) -> str:
        return f'MediaPoolItem(Name:"{self.name})"'



import sys
from typing import Any

from loguru import logger

logger.remove()


STR_FORMAT = (
    "<green>{time:YYYY-MM-DD HH:mm:ss}</green> | "
    "<level>{level: <8}</level> | "
    "<cyan>{module}</cyan>:<cyan>{line}</cyan> | <level>{message}</level> |"
)

logger.add(sink=sys.stderr, backtrace=True, format=STR_FORMAT)


def info(message: str, depth: int = 2) -> None:
    return logger.opt(record=True, depth=depth).info(message)


def warn(message: str, depth: int = 2) -> None:
    return logger.opt(record=True, depth=depth).warning(message)


def debug(message: str, depth: int = 2) -> None:
    return logger.opt(record=True, depth=depth).debug(message)


def error(message: str, depth: int = 2) -> None:
    return logger.opt(depth=depth).error(message)


def raise_exception(exception: Any, message: str, depth: int = 2) -> None:
    try:
        raise exception
    except exception:
        return logger.opt(exception=exception, depth=depth).info(message)



from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union

from typing_extensions import Literal, TypeAlias, TypedDict

import pydavinci.logger as log

if TYPE_CHECKING:
    from pydavinci.wrappers._resolve_stubs import (
        PyRemoteMediaPoolItem,
        PyRemoteTimeline,
        PyRemoteTimelineItem,
    )
    from pydavinci.wrappers.mediapoolitem import MediaPoolItem
    from pydavinci.wrappers.timeline import Timeline
    from pydavinci.wrappers.timelineitem import TimelineItem

    PydavinciParent: TypeAlias = Union[TimelineItem, Timeline, MediaPoolItem]

    RemoteMarkerParent: TypeAlias = Union[
        PyRemoteMediaPoolItem, PyRemoteTimeline, PyRemoteTimelineItem
    ]


class _RemoteMarkerData(TypedDict):
    name: str
    color: str
    note: str
    duration: int
    customData: str


RemoteMarkerData: TypeAlias = Dict[int, "_RemoteMarkerData"]


class MarkerData(TypedDict):
    frameid: int
    name: str
    color: str
    note: str
    duration: int
    customdata: str


ATTRS = Literal["frameid", "customdata", "color", "name", "duration", "note"]

COLORS = Literal[
    "Blue",
    "Cyan",
    "Green",
    "Yellow",
    "Red",
    "Pink",
    "Purple",
    "Fuchsia",
    "Rose",
    "Lavender",
    "Sky",
    "Mint",
    "Lemon",
    "Sand",
    "Cocoa",
    "Cream",
]


class MarkerCollection(object):
    def __init__(self, obj: "PydavinciParent") -> None:
        self._obj: "PydavinciParent" = obj
        self._parent_obj: "RemoteMarkerParent" = obj._obj
        self._cache: Dict[int, Marker] = {}
        self.fetch()

    def add(
        self,
        frameid: int,
        color: str,
        name: str,
        *,
        note: str = "",
        duration: int = 1,
        customdata: str = "",
        overwrite: bool = False,
    ) -> Optional["Marker"]:
        """
        Adds a marker.

        ``customdata`` is a ``str`` that can be used for programatically
        setting and searching for markers. It's not exposed to the GUI.

        Args:
            frameid (int): frame for marker to be inserted at
            color (str): marker color
            name (str): marker name
            note (str, optional): marker note. Defaults to empty.
            duration (int, optional): marker duration. Defaults to 1 frame.
            customdata (str, optional): custom user data. Defaults to empty.
            overwrite (bool, optional): set to `True` if you want to overwrite an existing marker at that frameid.

        Returns:
            bool: `Marker` if successful
        """
        # return self._parent_obj.AddMarker(frameid, color, name, note, duration, customdata)

        if frameid in self._cache:
            if not overwrite:
                log.info(
                    f"Marker at {frameid} already exists. Skipping... If you want to overwrite, use overwrite = True"
                )
                return None
            else:
                log.warn(f"Marker at frame {frameid} already exists. Overwriting ...")
                f = frameid
                self.delete(frameid=f)

        if self._parent_obj.AddMarker(frameid, color, name, note, duration, customdata):

            data: "MarkerData" = {
                "frameid": frameid,
                "color": color,
                "duration": duration,
                "note": note,
                "name": name,
                "customdata": customdata,
            }
            marker = Marker(self, self._parent_obj, data, frameid)
            self._cache_add(marker)
            return marker
        else:
            # TODO: try to inform user which error happened,
            # check if frameid and duration are OK for this clip
            log.error(
                "Couldn't add marker. Make sure the frameid is correct and the duration isn't bigger than the clips' length."
            )
            return None

    def find(self, needle: str) -> Optional["Marker"]:
        """Finds the first marker that matches `needle` for the `Marker's` `note`, `name` or `customdata`

        Returns:
            (Marker): first marker found with matching query
        """

        for marker in self._cache.values():
            if needle == marker.note or needle == marker.name or needle == marker.customdata:
                return marker

        return None

    def find_all(self, needle: str) -> Optional[List["Marker"]]:
        """Finds all markers that match `needle` for the `Marker's` `note`, `name` or `customdata`

        Returns:
            (Optional[List[Marker]]): all markers found or if none found, returns `None`
        """
        _ret: List[Marker] = []

        for marker in self._cache.values():
            if needle == marker.note or needle == marker.name or needle == marker.customdata:
                _ret.append(marker)

        return _ret if _ret else None

    def get_custom(self, customdata: str) -> Dict[Any, Any]:
        """
        Gets custom marker by ``customdata``

        Args:
            customdata (str): custom data string

        Returns:
            dict: dict with marker data
        """
        return self._parent_obj.GetMarkerByCustomData(customdata)

    def delete(
        self,
        *,
        frameid: int = 0,
        color: str = "",
        customdata: str = "",
    ) -> bool:
        """
        Deletes marker using ``frameid``, ``color`` or ``customdata``

        Args:
            frameid (int, optional): frameid to use for choosing which markers to delete
            color (str, optional): color to use for choosing which markers to delete
            customdata (str, optional): custom data to use for choosing which markers to delete

        Raises:
            ValueError: no valid params provided

        Returns:
            bool: ``True`` if successful, ``False`` otherwise

        Deleting Markers:
            When selecting by ``frameid``, will delete single marker

            When selecting by ``color``, will delete _all_ markers with provided color

            When selecting by ``customdata``, will delete first marker with matching custom data
        """

        if frameid:
            del self._cache[frameid]
            return self._parent_obj.DeleteMarkerAtFrame(frameid)
        elif color:
            # Delete all entries in the cache that are not the specified color
            # DeleteMarkersByColor() deletes all with that specified color
            self._cache = {
                frame: marker for frame, marker in self._cache.items() if marker.color != color
            }
            return self._parent_obj.DeleteMarkersByColor(color)

        elif customdata:
            # DeleteMarkerByCustomData deletes the first frame entry with the specified customdata,
            # so we're grabbing the keys and sorting them to ensure we delete the right entry in our
            # cache

            sorted = list(self._cache.keys())
            sorted.sort()
            self._parent_obj.DeleteMarkerByCustomData(customdata)
            del self._cache[sorted[0]]
            return True

        raise ValueError("You need to provide either 'frameid', 'color' or 'customdata'")

    def delete_all(self) -> None:
        """Deletes all markers"""
        for marker in self.all:
            marker.delete()

    @property
    def all(self) -> List["Marker"]:
        """Returns a list with all `Marker`'s

        Returns:
            (List[Marker])
        """
        self.fetch()
        return [x for x in self._cache.values()]

    def _cache_add(self, marker: "Marker") -> None:
        self._cache.update({marker._frameid: marker})

    def _cache_del(self, marker: "Marker") -> None:
        del self._cache[marker.frameid]

    def __iter__(self):  # type: ignore
        self.fetch()
        cache = self._cache.copy()
        yield from cache.values()

    def __repr__(self) -> str:
        markers = [str(x.frameid) for x in self.all]
        if len(markers) <= 4:
            displaystr = ", ".join(markers)
            return f"Markers(frames: {displaystr})"
        else:
            display: List[str] = markers[:2] + markers[-2:]
            display.insert(2, "...")
            displaystr = ", ".join(display)
            return f"Markers(Frames: {displaystr})"

    def fetch(self) -> None:
        """
        Fetch all markers from Davinci Resolve and updates `MarkerCollection`s internal cache. You probably won't need to use this.

        Why not:
            You would only use this if during the middle of the script execution a user manually added a marker. Otherwise, a `MarkerCollection` knows about all the markers
            it has deleted, added or updated, and `.fetch() ` is run on the class initialization.

        """
        markers: RemoteMarkerData = self._parent_obj.GetMarkers()
        if markers:
            for frameid in markers:
                marker: MarkerData = {
                    "frameid": frameid,
                    "color": markers[frameid]["color"],
                    "duration": markers[frameid]["duration"],
                    "name": markers[frameid]["name"],
                    "customdata": markers[frameid]["customData"],
                    "note": markers[frameid]["note"],
                }
                self._cache_add(Marker(self, self._parent_obj, marker, frameid))

        return


class Marker(object):
    def __init__(
        self,
        interface: "MarkerCollection",
        parent: "RemoteMarkerParent",
        data: "MarkerData",
        frameid: int,
    ) -> None:
        pass
        self._frameid = frameid
        self._data = data
        self._parent_obj = parent
        self._interface = interface

    def delete(self) -> None:
        """Deletes this `Marker`"""
        self._interface.delete(frameid=self._frameid)
        return

    @property
    def frameid(self) -> int:
        """Gets or changes this `Marker`'s `frameid`"""
        return self._data["frameid"]

    @frameid.setter
    def frameid(self, frameid: int) -> None:
        self.delete()
        self._frameid = frameid
        self._update("frameid", frameid)

    @property
    def customdata(self) -> str:
        """Gets or changes this `Marker`'s `customdata`"""
        return self._data["customdata"]

    @customdata.setter
    def customdata(self, customdata: str) -> None:
        if self._parent_obj.UpdateMarkerCustomData(self._frameid, customdata):
            self._data["customdata"] = customdata
            self._interface._cache_add(self)

    @property
    def name(self) -> str:
        """Gets or changes this `Marker`'s `name`"""
        return self._data["name"]

    @name.setter
    def name(self, name: str) -> None:
        self.delete()
        self._update("name", name)

    @property
    def color(self) -> str:
        """Gets or changes this `Marker`'s `color`"""
        return self._data["color"]

    @color.setter
    def color(self, color: Literal[COLORS]) -> None:
        # we don't have get_items on 3.6,
        # TODO update when support Davinci v18 and Python > 3.7
        if color not in COLORS.__values__:  # type: ignore
            return
        self.delete()
        self._update("color", color)

    @property
    def duration(self) -> int:
        """Gets or changes this `Marker`'s `duration`"""
        return self._data["duration"]

    @duration.setter
    def duration(self, duration: int) -> None:
        self.delete()
        self._update("duration", duration)

    @property
    def note(self) -> str:
        """Gets or changes this `Marker`'s `note`"""
        return self._data["note"]

    @note.setter
    def note(self, note: str) -> None:
        self.delete()
        self._update("note", note)

    def _update(self, key: ATTRS, value: Union[str, int]) -> None:
        self._data[key] = value  # type: ignore
        self._interface.add(
            frameid=self._data["frameid"],
            color=self._data["color"],
            name=self._data["name"],
            note=self._data["note"],
            duration=self._data["duration"],
            customdata=self._data["customdata"],
        )
        return

    def __repr__(self) -> str:
        return f"Marker(Frame: {self.frameid}, custom: {self.customdata})'"



from typing import TYPE_CHECKING, Any, List

from pydavinci.main import resolve_obj
from pydavinci.wrappers.mediapoolitem import MediaPoolItem

if TYPE_CHECKING:
    from pydavinci.wrappers._resolve_stubs import PyRemoteMediaStorage


class MediaStorage(object):
    def __init__(self) -> None:
        self._obj: "PyRemoteMediaStorage" = resolve_obj.GetMediaStorage()

    @property
    def mounted_volumes(self) -> List[Any]:
        """
        Gets list of OS mounted volumes

        Returns:
            list: list of volumes
        """
        return self._obj.GetMountedVolumeList()

    def get_subfolders(self, folder_path: str) -> List[str]:
        """
        Gets subfolders on ``folder_path``

        Args:
            folder_path (str): parent folder

        Returns:
            List[str]: list of subfolders
        """
        return self._obj.GetSubFolderList(folder_path)

    def get_file_list(self, folder_path: str) -> List[str]:
        """
        Gets list of file on ``folder_path``

        Args:
            folder_path (str): parent folder

        Returns:
            List[str]: list of files in ``parent_folder``
        """
        return self._obj.GetFileList(folder_path)

    def reveal_in_storage(self, path: str) -> bool:
        """
        Opens ``path`` in media storage

        Args:
            path (str): path

        Returns:
            None: None
        """
        return self._obj.RevealInStorage(path)

    def add_clip_mattes(
        self, mediapool_item: "MediaPoolItem", paths: List[str], stereo_eye: str
    ) -> bool:
        """
        Adds clip mattes from ``path`` in ``mediapool_item``

        Args:
            mediapool_item (MediaPoolItem): a MediaPoolItem object
            paths (Union[List[str], str]): A list of path strings or a single path string

        Returns:
            bool: ``True`` if successful, ``False`` otherwise
        """
        return self._obj.AddClipMattesToMediaPool(mediapool_item._obj, paths, stereo_eye)

    def add_timelilne_mattes(self, paths: List[str]) -> List["MediaPoolItem"]:
        """
        Adds timeline mattes into media pool

        Args:
            paths (Union[List[str], str]): a list of path strings or a single path string to a timeline matte

        Returns:
            List[MediaPoolItem]: list of timeline mattes ``MediaPoolItem``s
        """
        return [MediaPoolItem(x) for x in self._obj.AddTimelineMattesToMediaPool(paths)]

    def addclips_to_mediapool(self, item: List[str]) -> List["MediaPoolItem"]:
        """
        Adds clips to media pool

        Args:
            item (List[str]): a list of media paths or folders to add to media pool

        Returns:
            List[MediaPoolItem]: list of added ``MediaPoolItem``s

        Info:
            Even if you only added one clip, the returned ``MediaPoolItem`` will be inside a list. Access it directly by slicing it with the ``0`` index.
        """
        # /TODO Weird bug. Even if doing return type Union[List[MediaPoolItem], MediaPoolItem],
        # it doesn't identify the case that a single instace of MediaPoolItem might be returned,
        # and so on the linter it can't access the properties like .name

        # Work around is have everything output as a list right now, user should
        # have to use [0] on the object returned if it's a single clip.

        # if type(item) == type([]) and len(item) > 1:
        #     objs_added = [MediaPoolItem(x) for x in self._obj.AddItemListToMediaPool(item)]
        #     return objs_added
        # elif type(item) == type([]) and len(item) == 1:
        #     return [MediaPoolItem(self._obj.AddItemListToMediaPool(item[0]))]
        # else:
        #     return [MediaPoolItem(self._obj.AddItemListToMediaPool(item)[0])]

        objs_added = [MediaPoolItem(x) for x in self._obj.AddItemListToMediaPool(item)]
        return objs_added



from typing import TYPE_CHECKING, Any, Optional

from pydavinci.main import resolve_obj

if TYPE_CHECKING:
    from pydavinci.wrappers._resolve_stubs import PyRemoteResolve
    from pydavinci.wrappers.mediapool import MediaPool
    from pydavinci.wrappers.mediastorage import MediaStorage
    from pydavinci.wrappers.project import Project
    from pydavinci.wrappers.projectmanager import ProjectManager
    from pydavinci.wrappers.timeline import Timeline


class Resolve(object):
    def __init__(self, headless: Optional[bool] = None, path: Optional[str] = None):

        # this check is slow AF, to be implemented when we have auto-launch resolve

        # if not process_active("resolve"):
        #     raise NotImplementedError(
        #         "Couldn't find Davinci Resolve. Please make sure it's running"
        #     )

        self.pages = ["media", "cut", "edit", "fusion", "color", "fairlight", "deliver"]
        """Available pages to switch to using [``Resolve.page``][pydavinci.wrappers.resolve.Resolve.page]"""
        self._obj: PyRemoteResolve = resolve_obj

    @property
    def project_manager(self) -> "ProjectManager":
        """Returns the
        [``ProjectManager``][pydavinci.wrappers.projectmanager.ProjectManager-attributes] object.

        Returns:
            (ProjectManager): Project Manager object
        """
        from pydavinci.wrappers.projectmanager import ProjectManager

        return ProjectManager()

    @property
    def media_storage(self) -> "MediaStorage":
        """Returns the
        [``MediaStorage``][pydavinci.wrappers.mediastorage.MediaStorage-attributes] object.

        Returns:
            (MediaStorage): Media Storage object
        """
        from pydavinci.wrappers.mediastorage import MediaStorage

        return MediaStorage()

    @property
    def media_pool(self) -> "MediaPool":
        """Returns the
        [``MediaPool``][pydavinci.wrappers.mediapool.MediaPool-attributes] object.
        Returns:
            (MediaPool): Media Storage object
        """
        from pydavinci.wrappers.mediapool import MediaPool

        return MediaPool()

    @property
    def project(self) -> "Project":
        """Returns the current active
        [``Project``][pydavinci.wrappers.project.Project-attributes] object.

        Returns:
            (Project): Project object
        """
        from pydavinci.wrappers.project import Project

        return Project(resolve_obj.GetProjectManager().GetCurrentProject())

    @property
    def active_timeline(self) -> "Timeline":
        """Returns the current active
        [``Timeline``][pydavinci.wrappers.timeline.Timeline-attributes] object.

        Returns:
            (Timeline): Timeline object
        """
        from pydavinci.wrappers.timeline import Timeline

        return Timeline()

    @property
    def page(self) -> str:
        """
        Gets or sets current [``Resolve Page``][pydavinci.wrappers.resolve.Resolve.pages]. Note that certain methods are only available when in the right page.

        Args:
            page (str): valid page

        Returns:
            None: None
        """
        return self._obj.GetCurrentPage()

    @page.setter
    def page(self, page: str) -> None:
        """
        Page Setter test

        Args:
            page (str): valid page

        Returns:
            None: None
        """
        if page in self.pages:
            self._obj.OpenPage(page)
        else:
            validpages = " ".join(map(str, self.pages))
            raise ValueError(f'"{page}" is not a valid page. Available pages are: {validpages}')

    @property
    def product_name(self) -> str:
        """
        Returns:
            str: product name
        """
        return self._obj.GetProductName()

    @property
    def version(self) -> str:
        """
        Returns:
            str: version
        """
        return self._obj.GetVersionString()

    @property
    def fusion(self) -> Any:
        """Returns the Fusion object.

        Fusion Object:
            This is object is the same as the regular Fusion API. You can call ``fusion.__dir__()``
            to see all available methods.

            See [The last Fusion API documentation for more details.](https://documents.blackmagicdesign.com/UserManuals/Fusion8_Scripting_Guide.pdf)
        Returns:
            (Fusion): Fusion object
        """
        return resolve_obj.Fusion()

    def load_layout(self, layout_name: str) -> bool:
        """
        Loads saved layout named ``layout_name``

        Args:
            layout_name (str): layout name

        Returns:
            bool: ``True`` if successful, ``False`` otherwise
        """
        return self._obj.LoadLayoutPreset(layout_name)

    def update_layout(self, layout_name: str) -> bool:
        """
        Updates current layout to ``layout_name``

        Args:
            layout_name (str): layout to be updated

        Returns:
            bool: ``True`` if successful, ``False`` otherwise
        """
        return self._obj.UpdateLayoutPreset(layout_name)

    def save_layout(self, layout_name: str) -> bool:
        """
        Saves current layout as ``layout_name``

        Args:
            layout_name (str): layout name

        Returns:
            bool: ``True`` if successful, ``False`` otherwise
        """
        return self._obj.SaveLayoutPreset(layout_name)

    def import_layout(self, path: str, layout_name: str) -> bool:
        """
        Import ``layout_name`` from ``path``

        Args:
            path (str): path to layout file
            layout_name (str): name to be imported as

        Returns:
            bool: ``True`` if successful, ``False`` otherwise
        """

        return self._obj.ImportLayoutPreset(path, layout_name)

    def quit(self) -> None:
        """
        Quits Davinci Resolve

        Returns:
            None: None
        """
        return self._obj.Quit()

    def __repr__(self) -> str:
        return f'Resolve(Page: "{self.page}")'



from typing import Any, List

import pydavinci.main

# import psutil


def get_resolveobjs(objs: List[Any]) -> List[Any]:
    return [x._obj for x in objs]


TRACK_TYPES = ["video", "audio", "subtitle"]
TRACK_ERROR = "Track type must be: 'video', 'audio', or 'subtitle"


default_resolve_install = {
    "win": r"C:\\Program Files\\Blackmagic Design\\DaVinci Resolve\\Resolve.exe",
    "mac": "/Applications/DaVinci Resolve/DaVinci Resolve.app/Contents/MacOS/Resolve",
    "linux": "",
}


def is_resolve_obj(obj: Any) -> bool:
    if type(obj) == type(pydavinci.main.resolve_obj):  # noqa: E721
        return True
    else:
        return False


# def get_proc_pid(name: str) -> Union[None, int]:
#     for proc in psutil.process_iter():
#         try:
#             pinfo = proc.as_dict(attrs=["pid", "name"])
#             if (
#                 pinfo["name"].lower() == name.lower()
#                 or pinfo["name"].lower() == f"{name.lower()}.exe"
#             ):
#                 return pinfo["pid"]  # type: ignore
#         except psutil.AccessDenied:
#             # System process
#             pass
#         except psutil.NoSuchProcess:
#             # Process terminated
#             pass
#     return None


# def process_active(process_name: str) -> bool:

#     if process_name or f"{process_name}.exe" in (p.name().lower() for p in psutil.process_iter()):
#         return True
#     return False


# def launch_resolve(headless: Optional[bool] = False, path: Optional[str] = None):  # type: ignore

#     if not get_proc_pid("resolve"):
#         system: str = ""
#         args: List[str] = []
#         if sys.platform.startswith("win32"):
#             system = "win"
#         elif sys.platform.startswith("darwin"):
#             system = "mac"
#         else:
#             system = "linux"

#         if path:
#             args.append(path)
#         else:
#             args.append(default_resolve_install[system])

#         if headless:
#             args.append("-nogui")

#         kwargs: Dict[Any, Any] = {}
#         # try:
#         if system == "win":
#             # Windows fuckery here to spawn a process without it being a child
#             # of the python interpreter
#             # https://docs.microsoft.com/pt-br/windows/win32/procthread/process-creation-flags?redirectedfrom=MSDN
#             # this holds the tests...
#             CREATE_NEW_PROCESS_GROUP = 0x00000200
#             DETACHED_PROCESS = 0x00000008
#             kwargs.update(creationflags=DETACHED_PROCESS | CREATE_NEW_PROCESS_GROUP)
#         # elif system == "wsl":
#         #     args.insert(0, "cmd.exe")
#         else:
#             kwargs.update(close_fds=True)

#         subprocess.Popen(
#             args,
#             start_new_session=True,
#             stdout=subprocess.DEVNULL,
#             stderr=subprocess.DEVNULL,
#             **kwargs,
#         )
#         print("subiu processo")
#         while not get_proc_pid("fuscript"):
#             print("Waiting for fuscript")

#         ready = False
#         while not ready:
#             time.sleep(0.3)
#             print("...")
#             try:
#                 print("try")
#                 if pydavinci.main.get_resolve().GetProjectManager() is not None:
#                     print(pydavinci.main.get_resolve().GetProjectManager())
#                     print("chego aqui")
#                     ready = True
#                     break
#             except AttributeError:
#                 continue

#         print("pronto pra vaza")
#         time.sleep(4)
#         return True

#         # except FileNotFoundError:
#         #     print("Davinci Resolve executable not found. Please double check the path")

#     print("Davinci Resolve already running... Continuing")



def load_fusionscript():  # type: ignore
    import imp
    import os
    import sys

    WIN_ENV_VARIABLES = {
        "RESOLVE_SCRIPT_API": r"%PROGRAMDATA%\Blackmagic Design\DaVinciResolve\Support\Developer\Scripting",
        "RESOLVE_SCRIPT_LIB": r"C:\Program Files\Blackmagic Design\DaVinciResolve\fusionscript.dll",
        "PYTHONPATH": r"%PYTHONPATH%;%RESOLVE_SCRIPT_API%\Modules\\",
    }

    MAC_ENV_VARIABLES = {
        "RESOLVE_SCRIPT_API": "/Library/Application Support/Blackmagic Design/DaVinciResolve/Developer/Scripting",
        "RESOLVE_SCRIPT_LIB": "/Applications/DaVinci Resolve/DaVinciResolve.app/Contents/Libraries/Fusion/fusionscript.so",
        "PYTHONPATH": "$PYTHONPATH:$RESOLVE_SCRIPT_API/Modules/",
    }
    LINUX_ENV_VARIABLES = {
        "RESOLVE_SCRIPT_API": "/opt/resolve/Developer/Scripting",
        "RESOLVE_SCRIPT_LIB": "/opt/resolve/libs/Fusion/fusionscript.so",
        "PYTHONPATH": "$PYTHONPATH:$RESOLVE_SCRIPT_API/Modules/",
    }

    if sys.platform.startswith("win32"):
        for key in WIN_ENV_VARIABLES.keys():
            os.environ[key] = WIN_ENV_VARIABLES[key]

    elif sys.platform.startswith("darwin"):
        for key in MAC_ENV_VARIABLES.keys():
            os.environ[key] = MAC_ENV_VARIABLES[key]

    else:
        for key in LINUX_ENV_VARIABLES.keys():
            os.environ[key] = LINUX_ENV_VARIABLES[key]

    script_module = None

    try:
        import fusionscript as script_module  # type: ignore
    except ImportError:
        # Look for installer based environment variables:
        import os

        lib_path = os.getenv("RESOLVE_SCRIPT_LIB")
        if lib_path:
            try:
                script_module = imp.load_dynamic("fusionscript", lib_path)
            except ImportError:
                pass
        if not script_module:
            # Look for default install locations:
            ext = ".so"
            if sys.platform.startswith("darwin"):
                path = (
                    "/Applications/DaVinci Resolve/DaVinci Resolve.app/Contents/Libraries/Fusion/"
                )
            elif sys.platform.startswith("win") or sys.platform.startswith("cygwin"):
                ext = ".dll"
                path = "C:\\Program Files\\Blackmagic Design\\DaVinci Resolve\\"
            else:
                path = "/opt/resolve/libs/Fusion/"

            try:
                script_module = imp.load_dynamic(
                    "fusionscript", f"{path}fusionscript{ext}"
                )  # noqa: E501, B950 # type: ignore
            except ImportError:
                pass



from typing import Optional

import pydavinci.logger as log


class ObjectNotFound(BaseException):
    pass


class TimelineNotFound(BaseException):
    def __init__(self, *args: object, extra: Optional[str] = None) -> None:
        self.message = "Couldn't find a valid timeline."
        if extra:
            log.error(extra)

        super().__init__(*args, self.message)



from typing import TYPE_CHECKING, List

from pydavinci.utils import is_resolve_obj

if TYPE_CHECKING:
    from pydavinci.wrappers._resolve_stubs import PyRemoteFolder
from pydavinci.wrappers.mediapoolitem import MediaPoolItem


class Folder(object):
    def __init__(self, obj: "PyRemoteFolder") -> None:
        if is_resolve_obj(obj):
            self._obj: "PyRemoteFolder" = obj
        else:
            raise TypeError(f"{type(obj)} is not a valid {self.__class__.__name__} type")

    @property
    def clips(self) -> List[MediaPoolItem]:
        """
        Gets all clips

        Returns:
            (List[MediaPoolItem]): list of clips
        """
        from pydavinci.wrappers.mediapoolitem import MediaPoolItem

        objs = self._obj.GetClipList()
        return [MediaPoolItem(x) for x in objs]

    @property
    def name(self) -> str:
        """
        Gets folder name

        Returns:
            name(str): folder name
        """
        return self._obj.GetName()

    @property
    def subfolders(self) -> List["Folder"]:
        """
        Gets subfolders

        Returns:
            subfolders(List[Folder]): subfolders
        """
        return [Folder(x) for x in self._obj.GetSubFolderList()]

    def __repr__(self) -> str:
        return f'Folder(Name:"{self.name})"'



from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union

import pydavinci.logger as log
from pydavinci.exceptions import TimelineNotFound
from pydavinci.main import resolve_obj
from pydavinci.utils import TRACK_ERROR, TRACK_TYPES, get_resolveobjs, is_resolve_obj
from pydavinci.wrappers.marker import MarkerCollection
from pydavinci.wrappers.settings.constructor import get_tl_settings
from pydavinci.wrappers.timelineitem import TimelineItem


if TYPE_CHECKING:
    from pydavinci.wrappers._resolve_stubs import PyRemoteTimeline
    from pydavinci.wrappers.settings.constructor import TimelineSettings


class Timeline(object):
    def __init__(self, *args: Any) -> None:
        if args:
            if is_resolve_obj(args[0]):
                self._obj: "PyRemoteTimeline" = args[0]
            else:
                raise TypeError(f"{type(args[0])} is not a valid {self.__class__.__name__} type")

        else:
            _obj = resolve_obj.GetProjectManager().GetCurrentProject().GetCurrentTimeline()
            if _obj:
                self._obj = _obj
            else:
                raise TimelineNotFound(
                    extra="Couldn't find any active timeline. Are you sure there's any timeline in the project?"
                )

        self.markers = MarkerCollection(self)
        self._settings: Optional[TimelineSettings] = None

    def custom_settings(self, use: bool) -> bool:
        # Davinci only allows setting timeline settings if "useCustomSettings" is true, otherwise it returns False every time.
        """Allows this timeline to have settings independent from the project settings. See [Quickstart on Settings](../settings#project-vs-timeline-settings) for more details.

        Args:
            use (bool): ``True`` to use independent settings, ``False`` to follow project settings.

        """
        if use:
            return self.set_setting("useCustomSettings", "1")
        else:
            return self.set_setting("useCustomSettings", "0")

    @property
    def settings(self) -> "TimelineSettings":
        """Returns the [`TimelineSettings`](../settings/timeline) interface.
        [`Timeline.custom_settings(True)`](./#pydavinci.wrappers.timeline.Timeline.custom_settings) must be called first."""
        if self.get_setting("useCustomSettings") == "0":
            # doing the check here again in case user uses self.set_setting("useCustomSettings")
            # need to be compatible with that too

            log.error(
                "Can't create timeline settings. Timeline not configured for custom settings. "
                + "Use Timeline.custom_settings(True) and then call Timeline.settings again."  # noqa: W503
            )

            return  # type: ignore

        if self._settings:
            return self._settings
        else:
            self._settings = get_tl_settings(self)
            return self._settings

    @property
    def name(self) -> str:
        """
        Gets or sets timeline ``name``

        Args:
            name (str): new timeline name

        Returns:
            bool: ``True`` if successful, ``False`` otherwise
        """
        return self._obj.GetName()

    @name.setter
    def name(self, name: str) -> bool:
        """
        Changes timeline to ``name``

        Args:
            name (str): new timeline name

        Returns:
            bool: ``True`` if successful, ``False`` otherwise
        """
        return self._obj.SetName(name)

    def activate(self) -> bool:
        """
        Makes this timeline active

        Returns:
            bool: ``True`` if successful, ``False`` otherwise
        """
        return resolve_obj.GetProjectManager().GetCurrentProject().SetCurrentTimeline(self._obj)

    @property
    def start_frame(self) -> int:
        """
        Gets timeline start frame

        Returns:
            int: start frame
        """
        return self._obj.GetStartFrame()

    @property
    def end_frame(self) -> int:
        """
        Gets timeline end frame

        Returns:
            int: end frame
        """
        return self._obj.GetEndFrame()

    def track_count(self, track_type: str) -> int:
        """
        Gets track count on ``track_type``

        Args:
            track_type (str): valid ``track_types``: ``video``, ``audio``, ``subtitle``

        Raises:
            ValueError: Not a valid track type

        Returns:
            int: number of tracks
        """
        if track_type not in TRACK_TYPES:
            raise ValueError(TRACK_ERROR)
        else:
            return self._obj.GetTrackCount(track_type)

    def items(self, track_type: str, track_index: int) -> List["TimelineItem"]:
        """
        Gets [``TimelineItem``][pydavinci.wrappers.timelineitem.TimelineItem-attributes]s from a track

        Args:
            track_type (str): valid ``track_type``: ``video``, ``audio``, ``subtitle``
            track_index (int): track index. Starts at ``1``

        Raises:
            ValueError: Not a valid track type

        Returns:
            (List[TimelineItem]): list of items at specified track
        """
        if track_type not in TRACK_TYPES:
            raise ValueError(TRACK_ERROR)

        return [TimelineItem(x) for x in self._obj.GetItemListInTrack(track_type, track_index)]

    def apply_grade_from_DRX(
        self, drx_path: str, grade_mode: int, timeline_items: List["TimelineItem"]
    ) -> bool:
        """
        Applies drx grade on a list of
        [``TimelineItem``][pydavinci.wrappers.timelineitem.TimelineItem-attributes]s

        Args:
            drx_path (str): path to a ``.drx`` file
            grade_mode (int): grade mode to use. ``0`` for ``No Keyframes``, ``1`` for ``Source Timecode aligned`` and ``2`` for ``Start Frames aligned``.
            timeline_items (List[TimelineItem]): timeline items to apply grade to

        Returns:
            bool: ``True`` if successful, ``False`` otherwise
        """
        return self._obj.ApplyGradeFromDRX(drx_path, grade_mode, [x._obj for x in timeline_items])

    @property
    def timecode(self) -> str:
        """
        Gets current timecode

        Returns:
            str: timecode
        """
        return self._obj.GetCurrentTimecode()

    @property
    def current_video_item(self) -> "TimelineItem":
        """
        Returns current video ``TimelineItem``

        Returns:
            TimelineItem: current ``TimelineItem``
        """
        if resolve_obj.GetCurrentPage() != "edit":
            raise Warning("You need to switch to edit page first before getting using this method.")
        return TimelineItem(self._obj.GetCurrentVideoItem())

    @property
    def current_clip_thumbnail(self) -> Dict[Any, Any]:
        """
        Returns a dict with data containing metadata + raw thumbnail
        image data (RGB 8-bit image data encoded in base64 format) for current media in the Color Page.

        Returns:
            dict: (keys "width", "height", "format" and "data")
        """
        return self._obj.GetCurrentClipThumbnailImage()

    def get_track_name(self, track_type: str, track_index: int) -> str:
        """
        Gets track name

        Args:
            track_type (str): valid track_type: ``video``, ``audio`` or ``subtitle``
            track_index (int): track index. Starts at ``1``

        Raises:
            ValueError: Not a valid track type

        Returns:
            str: track name
        """
        if track_type not in TRACK_TYPES:
            raise ValueError(TRACK_ERROR)
        return self._obj.GetTrackName(track_type, track_index)

    def set_track_name(self, track_type: str, track_index: int, new_name: str) -> bool:
        """
        Sets track name

        Args:
            track_type (str): valid track_type: ``video``, ``audio`` or ``subtitle``
            track_index (int): track index. Starts at ``1``
            new_name (str): new name

        Raises:
            ValueError: Not a valid track type

        Returns:
            (bool): ``True`` if successful, ``False`` otherwise
        """
        if track_type not in TRACK_TYPES:
            raise ValueError(TRACK_ERROR)
        return self._obj.SetTrackName(track_type, track_index, new_name)

    def duplicate_timeline(self, timeline_name: Optional[str] = None) -> "Timeline":
        """
        Duplicates this timeline

        Args:
            timeline_name (str, optional): New timeline name. If not provided, appends "Copy" to timeline name.

        Returns:
            (Timeline): new duplicated timeline
        """
        if timeline_name:
            return Timeline(self._obj.DuplicateTimeline(timeline_name))
        else:
            return Timeline(self._obj.DuplicateTimeline())

    def create_compound_clip(
        self, timeline_items: List["TimelineItem"], clip_info: Optional[Dict[Any, Any]] = None
    ) -> "TimelineItem":
        """
        Creates a compound clip using ``timeline_items``

        Args:
            timeline_items (List[TimelineItem]): list of ``TimelineItem``s to use
            clip_info (dict, optional): optional compound clip settings. Valid dict:
                ```python
                {"startTimecode" : "00:00:00:00", "name" : "Compound Clip 1"}
                ```

        Returns:
            (TimelineItem): compound clip
        """
        if not clip_info:
            return TimelineItem(self._obj.CreateCompoundClip(get_resolveobjs(timeline_items)))

        return TimelineItem(
            self._obj.CreateCompoundClip(get_resolveobjs(timeline_items), clip_info)
        )

    def create_fusion_clip(self, timeline_items: List["TimelineItem"]) -> "TimelineItem":
        """
        Creates a Fusion clip with ``timeline_items``

        Args:
            timeline_items (List[TimelineItem]): timeline items to be used as input in the Fusion clip

        Returns:
            (TimelineItem): resulting fusion clip
        """
        return TimelineItem(self._obj.CreateFusionClip(get_resolveobjs(timeline_items)))

    def import_aaf_into_timeline(
        self, file_path: str, import_options: Optional[Dict[Any, Any]] = None
    ) -> bool:  # noqa: E501
        """
        Imports an aaf into the timeline

        Import Options:
            Imports timeline items from an AAF file and optional ``import_options`` dict into the timeline, with support for the keys:

            ``"autoImportSourceClipsIntoMediaPool"``: ``bool``, specifies if source clips should be imported into media pool, ``True`` by default

            ``"ignoreFileExtensionsWhenMatching"``: ``bool``, specifies if file extensions should be ignored when matching, ``False`` by default

            ``"linkToSourceCameraFiles"``: bool, specifies if link to source camera files should be enabled, False by default

            ``"useSizingInfo"``: ``bool``, specifies if sizing information should be used, ``False`` by default

            ``"importMultiChannelAudioTracksAsLinkedGroups"``: bool, specifies if multi-channel audio tracks should be imported as linked groups, ``False`` by default

            ``"insertAdditionalTracks"``: bool, specifies if additional tracks should be inserted, ``True`` by default

            ``"insertWithOffset"``: ``str``, specifies insert with offset value in timecode format - defaults to ``"00:00:00:00"``, applicable if "insertAdditionalTracks" is False

            ``"sourceClipsPath"``: ``str``, specifies a filesystem path to search for source clips if the media is inaccessible in their original path and if ``"ignoreFileExtensionsWhenMatching"`` is ``True``

            ``"sourceClipsFolders"``: ``str``, list of ``Folder`` objects to search for source clips if the media is not present in current folder

        Args:
             file_path (str): path to ``.aaf`` file
             import_options (dict, optional): optional import options. See description above. Defaults to {}.

        Returns:
             bool: ``True`` if successful, ``False`` otherwise

        """
        if not import_options:
            return self._obj.ImportIntoTimeline(file_path)
        return self._obj.ImportIntoTimeline(file_path, import_options)

    def export(
        self, file_name: str, export_type: str, export_subtype: Optional[str] = None
    ) -> bool:
        """
        Exports timeline file ``(.aaf, .xml, etc)``
        Supported ``export_type``:
        ```
        "AAF"
        "DRT"
        "EDL"
        "FCP_7_XML"
        "FCPXML_1_3"
        "FCPXML_1_4"
        "FCPXML_1_5"
        "FCPXML_1_6"
        "FCPXML_1_7"
        "FCPXML_1_8"
        "HDR_10_PROFILE_A"
        "HDR_10_PROFILE_B"
        "TEXT_CSV"
        "TEXT_TAB"
        "DOLBY_VISION_VER_2_9"
        "DOLBY_VISION_VER_4_0"
        ```

        Supported ``export_subtype``:
        ```python
        "NONE"
        "AAF_NEW"
        "AAF_EXISTING"
        "CDL"
        "SDL"
        "MISSING_CLIPS"
        ```

        Export types and subtypes:
            Please note that ``export_subtype`` is a required parameter for ``AAF`` and ``EDL``. For rest of the ``export_type``, ``export_subtype`` is ignored.

            When ``export_type`` is ``AAF``, valid ``export_subtype`` values are ``AAF_NEW`` and ``AAF_EXISTING``.

            When ``export_type`` is ``EXPORT_EDL``, valid export_dubtype values are ``EXPORT_CDL``, ``EXPORT_SDL``, ``EXPORT_MISSING_CLIPS`` and ``EXPORT_NONE``.

        Args:
            file_name (str): full filepath to export to including file name
            export_type (_type_): supported export type
            export_subtype (_type_): supported export subtype

        Returns:
            bool: ``True`` if successful, ``False`` otherwise
        """
        export_type_map = {
            "AAF": 0.0,
            "DRT": 1.0,
            "EDL": 2.0,
            "FCP_7_XML": 3.0,
            "FCPXML_1_3": 4.0,
            "FCPXML_1_4": 5.0,
            "FCPXML_1_5": 6.0,
            "FCPXML_1_6": 7.0,
            "FCPXML_1_7": 8.0,
            "FCPXML_1_8": 9.0,
            "HDR_10_PROFILE_A": 10.0,
            "HDR_10_PROFILE_B": 11.0,
            "TEXT_CSV": 12.0,
            "TEXT_TAB": 13.0,
            "DOLBY_VISION_VER_2_9": 14.0,
            "DOLBY_VISION_VER_4_0": 15.0,
        }

        export_subtype_map = {
            "NONE": -1.0,
            "AAF_NEW": 0.0,
            "AAF_EXISTING": 1.0,
            "CDL": 2.0,
            "SDL": 3.0,
            "MISSING_CLIPS": 4.0,
        }

        if export_subtype:

            return self._obj.Export(
                file_name, export_type_map[export_type], export_subtype_map[export_subtype]
            )
        else:
            return self._obj.Export(file_name, export_type_map[export_type])

    def get_setting(
        self, settingname: Optional[str] = None
    ) -> Union[str, int, float, Dict[Any, Any]]:
        """
        _This function is a fallback if using [`Timeline.settings`][pydavinci.wrappers.timeline.Timeline.settings] doesn't work._

        Get timeline setting. If no setting provided, returns a dict with all settings.

        Args:
            settingname (str, optional): setting name.

        Returns:
            Union[str, Dict]: setting(s)
        """
        if settingname:
            return self._obj.GetSetting(settingname)
        return self._obj.GetSetting()

    def set_setting(self, setting_name: str, value: Union[str, int, float, Dict[Any, Any]]) -> bool:
        """
        _This function is a fallback if using [`Timeline.settings`][pydavinci.wrappers.timeline.Timeline.settings] doesn't work._

        Sets `Timeline` setting ``seting_name`` to ``value``

        Args:
            setting_name (str): setting name
            value (Union[str, int, Dict[Any, Any]): setting value

        Returns:
           ``True`` if successful, ``False`` otherwise
        """
        return self._obj.SetSetting(setting_name, value)

    def insert_generator(self, generator_name: str) -> "TimelineItem":
        """
        Inserts a generator in the timeline

        Args:
            generator_name (str): generator name to be inserted

        Returns:
            (TimelineItem): generator
        """
        return TimelineItem(self._obj.InsertGeneratorIntoTimeline(generator_name))

    def insert_fusion_generator(self, generator_name: str) -> "TimelineItem":
        """
        Inserts a fusion generator in the timeline

        Args:
            generator_name (str): fusion generator name

        Returns:
            (TimelineItem): fusion generator
        """
        return TimelineItem(self._obj.InsertFusionGeneratorIntoTimeline(generator_name))

    def insert_ofx_generator(self, generator_name: str) -> "TimelineItem":
        """
        Inserts an OFX generator in the timeline

        Args:
            generator_name (str): OFX generator name

        Returns:
            (TimelineItem): OFX generator
        """
        return TimelineItem(self._obj.InsertOFXGeneratorIntoTimeline(generator_name))

    def insert_title(self, title_name: str) -> "TimelineItem":
        """
        Inserts a title in the timeline

        Args:
            title_name (str): title name

        Returns:
            (TimelineItem): title
        """
        return TimelineItem(self._obj.InsertTitleIntoTimeline(title_name))

    def insert_fusion_title(self, title_name: str) -> "TimelineItem":
        """
        Inserts a fusion title in the timeline

        Args:
            title_name (str): fusion title name

        Returns:
            (TimelineItem): fusion title
        """
        return TimelineItem(self._obj.InsertFusionTitleIntoTimeline(title_name))

    def __repr__(self) -> str:
        return f"Timeline(name: {self.name})"



from pathlib import Path
from typing import Optional, Union

from pydantic import DirectoryPath


def bool_to_intstr(bool: bool) -> str:
    return str(int(bool))


def none_to_str(value: None = None) -> str:
    if value is None:
        return ""
    return value


def path_to_str(value: Union[Path, DirectoryPath, str]) -> Optional[str]:
    if isinstance(value, Path) or isinstance(value, DirectoryPath):
        return str(value)
    elif value == "":
        return None
    return value


def super_scale_transform(scale: Union[int, str]) -> Optional[Union[str, int]]:
    if isinstance(scale, int):
        if scale == 0:
            return "auto"
        elif scale == 1:
            return "no_scaling"
        elif scale == 2:
            return "2x"
        elif scale == 3:
            return "3x"
        elif scale == 4:
            return "4x"
    else:
        if scale == "auto":
            return 0
        elif scale == "no_scaling":
            return 1
        elif scale == "2x":
            return 2
        elif scale == "3x":
            return 3
        elif scale == "4x":
            return 4

    return scale


# This is a map that's applied when going from pydantic to the Resolve API.
# Bool to intstr for example gets the bool type from Python, and converts it
# either to a string "1" or "0"

# Superscale has it's own shenanigans

# Also noted are some observations I made while testing out some settings.
# Didn't test them all cause I'm not that masochist.


SETTINGS_MAP = {
    "audioCaptureNumChannels": str,
    "audioOutputHasTimecode": bool_to_intstr,
    "audioPlayoutNumChannels": str,
    "colorAcesGamutCompressType": str,
    "colorAcesIDT": str,
    "colorAcesNodeLUTProcessingSpace": str,
    "colorAcesODT": str,
    "colorGalleryStillsLocation": path_to_str,
    "colorGalleryStillsNamingCustomPattern": none_to_str,
    "colorGalleryStillsNamingEnabled": bool_to_intstr,
    "colorGalleryStillsNamingPattern": str,
    "colorGalleryStillsNamingWithStillNumber": str,
    "colorKeyframeDynamicsEndProfile": str,
    "colorKeyframeDynamicsStartProfile": str,
    "colorLuminanceMixerDefaultZero": bool_to_intstr,
    "colorScienceMode": str,
    "colorSpaceInput": str,
    "colorSpaceInputGamma": none_to_str,
    "colorSpaceOutput": str,
    "colorSpaceOutputGamma": none_to_str,
    "colorSpaceOutputGamutMapping": str,
    "colorSpaceOutputGamutSaturationKnee": str,
    "colorSpaceOutputGamutSaturationMax": str,
    "colorSpaceOutputToneLuminanceMax": str,
    "colorSpaceOutputToneMapping": str,
    "colorSpaceTimeline": str,
    "colorSpaceTimelineGamma": none_to_str,
    "colorUseBGRPixelOrderForDPX": bool_to_intstr,
    "colorUseContrastSCurve": bool_to_intstr,
    "colorUseLegacyLogGrades": str,
    "colorUseLocalVersionsAsDefault": bool_to_intstr,
    "colorUseStereoConvergenceForEffects": bool_to_intstr,
    "colorVersion10Name": none_to_str,
    "colorVersion1Name": none_to_str,
    "colorVersion2Name": none_to_str,
    "colorVersion3Name": none_to_str,
    "colorVersion4Name": none_to_str,
    "colorVersion5Name": none_to_str,
    "colorVersion6Name": none_to_str,
    "colorVersion7Name": none_to_str,
    "colorVersion8Name": none_to_str,
    "colorVersion9Name": none_to_str,
    "graphicsWhiteLevel": str,
    "hdr10PlusControlsOn": bool_to_intstr,
    "hdrDolbyControlsOn": bool_to_intstr,
    "hdrDolbyMasterDisplay": none_to_str,
    "hdrDolbyVersion": str,
    "hdrMasteringLuminanceMax": str,
    "hdrMasteringOn": bool_to_intstr,
    "imageDeinterlaceQuality": str,
    "imageEnableFieldProcessing": bool_to_intstr,
    "imageMotionEstimationMode": str,
    "imageMotionEstimationRange": str,
    "imageResizeMode": str,
    "imageResizingGamma": str,
    "imageRetimeInterpolation": str,
    "inputDRT": str,
    "inputDRTSatRolloffLimit": str,
    "inputDRTSatRolloffStart": str,
    "isAutoColorManage": str,
    "limitAudioMeterAlignLevel": str,
    "limitAudioMeterDisplayMode": str,
    "limitAudioMeterHighLevel": str,
    "limitAudioMeterLUFS": str,
    "limitAudioMeterLoudnessScale": str,
    "limitAudioMeterLowLevel": str,
    "limitBroadcastSafeLevels": str,
    "limitBroadcastSafeOn": bool_to_intstr,
    "limitSubtitleCPL": str,
    "limitSubtitleCaptionDurationSec": str,
    "outputDRT": str,
    "outputDRTSatRolloffLimit": str,
    "outputDRTSatRolloffStart": str,
    "perfAutoRenderCacheAfterTime": str,
    "perfAutoRenderCacheComposite": bool_to_intstr,
    "perfAutoRenderCacheEnable": bool_to_intstr,
    "perfAutoRenderCacheFuEffect": bool_to_intstr,
    "perfAutoRenderCacheTransition": str,
    "perfCacheClipsLocation": path_to_str,
    "perfOptimisedCodec": str,
    "perfOptimisedMediaOn": bool_to_intstr,  # Menu: Playback -> Use Optimized media if Available
    "perfOptimizedResolutionRatio": str,
    "perfProxyMediaOn": bool_to_intstr,  # Menu: Playback -> Use Proxy Media NOTE: this might change with the new v18 - test
    "perfProxyResolutionRatio": str,  # This is not on the project settings, it's on the menu Playback -> Timeline Proxy Mode
    "perfRenderCacheCodec": str,
    "perfRenderCacheMode": str,  # Menu: Playback -> Render cache mode
    "rcmPresetMode": str,
    "separateColorSpaceAndGamma": bool_to_intstr,
    "superScaleNoiseReduction": str,
    "superScaleSharpness": str,
    "timelineDropFrameTimecode": bool_to_intstr,
    "timelineFrameRate": str,  # only allows setting in str and if there's no media already in the project, otherwise need to use custom timeline settings for it to work, and there's a lot of weird stuff where you can't set the timeline settings.... Probably best to, before importing media, set the timeline framerate you want...
    "timelineFrameRateMismatchBehavior": str,
    "timelineInputResMismatchBehavior": str,
    "timelineInputResMismatchCustomPreset": str,  # It has to be in Image Scaling -> Resize Filter: Custom, but can't seem to be able to set any custom presets.
    "timelineInputResMismatchUseCustomPreset": str,
    "timelineInterlaceProcessing": bool_to_intstr,
    "timelineOutputPixelAspectRatio": str,
    "timelineOutputResMatchTimelineRes": bool_to_intstr,
    "timelineOutputResMismatchBehavior": str,
    "timelineOutputResMismatchCustomPreset": str,
    "timelineOutputResMismatchUseCustomPreset": bool_to_intstr,
    "timelineOutputResolutionHeight": str,
    "timelineOutputResolutionWidth": str,
    "timelinePixelAspectRatio": str,
    "timelinePlaybackFrameRate": str,  # don't know why, but I can't seem to set this using either str or float
    "timelineResolutionHeight": str,
    "timelineResolutionWidth": str,
    "timelineSaveThumbsInProject": bool_to_intstr,
    "timelineWorkingLuminance": str,  # couldn't find it
    "timelineWorkingLuminanceMode": str,  # couldn't find it
    "useCATransform": bool_to_intstr,
    "useColorSpaceAwareGradingTools": bool_to_intstr,
    "useInverseDRT": bool_to_intstr,
    "videoCaptureCodec": none_to_str,
    "videoCaptureFormat": none_to_str,
    "videoCaptureIngestHandles": str,
    "videoCaptureLocation": path_to_str,
    "videoCaptureMode": str,
    "videoDataLevels": str,  # BUG: Apparently doesn't work.. doesn't change the setting
    "videoDataLevelsRetainSubblockAndSuperWhiteData": bool_to_intstr,
    "videoDeckAdd32Pulldown": bool_to_intstr,
    "videoDeckBitDepth": str,
    "videoDeckFormat": str,
    "videoDeckNonAutoEditFrames": bool_to_intstr,
    "videoDeckOutputSyncSource": str,
    "videoDeckPrerollSec": str,
    "videoDeckSDIConfiguration": str,  # BUG: Only works with single_link, dual and quad link return "none". Cant set it too
    "videoDeckUse444SDI": bool_to_intstr,
    "videoDeckUseAudoEdit": bool_to_intstr,  # BUG: Typo - should be "auto"
    "videoDeckUseStereoSDI": bool_to_intstr,
    "videoMonitorBitDepth": str,
    "videoMonitorFormat": str,
    "videoMonitorMatrixOverrideFor422SDI": str,
    "videoMonitorSDIConfiguration": str,
    "videoMonitorScaling": str,
    "videoMonitorUse444SDI": bool_to_intstr,
    "videoMonitorUseHDROverHDMI": bool_to_intstr,
    "videoMonitorUseLevelA": bool_to_intstr,
    "videoMonitorUseMatrixOverrideFor422SDI": bool_to_intstr,
    "videoMonitorUseStereoSDI": bool_to_intstr,
    "videoPlayoutAudioFramesOffset": int,
    "videoPlayoutBatchHeadDuration": str,
    "videoPlayoutBatchTailDuration": str,
    "videoPlayoutLTCFramesOffset": str,
    "videoPlayoutMode": str,
    "videoPlayoutShowLTC": bool_to_intstr,
    "videoPlayoutShowSourceTimecode": bool_to_intstr,
    "superScale": super_scale_transform,
    "useCustomSettings": bool_to_intstr,
    "selfvalidate": bool,
}
